/*
    This file is part of the FElt finite element analysis package.
    Copyright (C) 1993,1994 Jason I. Gobat and Darren C. Atkinson

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/****************************************************************************
*
* File:         felt.c
*
* Description:  Contains code for the driver application for the Finite 
*               ELemenT (FELT) package currently under development.
*
*****************************************************************************/

# include <stdio.h>
# include <string.h>
# include "problem.h"
# include "fe.h"
# include "error.h"
# include "allocate.h"

# define streq(a,b)	!strcmp(a,b)


static char *usage = "\
usage: felt [options] [filename]\n\
       -debug              write debugging output\n\
       +table              do not print tabular dynamic results\n\
       -plot               do time-displacement plot for dynamic results\n\
       -renumber           force automatic node renumbering\n\
       -summary            include material summary statistics\n\
       -printk             print the global stiffness matrix\n\
       -graphics filename  create file for structure visualization\n\
       -nocpp              do not use a preprocessor\n\
       -cpp filename       preprocessor to use\n\
       -Dname[=value]      define a macro\n\
       -Uname              undefine a macro\n\
       -Idirectory         specify include directory\n\
";


static int   debug = 0;
static int   summary = 0;
static int   printk = 0;
static int   doplot = 0;
static int   dotable = 1;
static int   renumber = 0;
static char *graphics = NULL;


/************************************************************************
 * Function:	ParseFeltOptions					*
 *									*
 * Description:	Parses the felt specific command line options.		*
 ************************************************************************/

static int ParseFeltOptions (argc, argv)
    int  *argc;
    char *argv [ ];
{
    int   i;
    int   j;
    char *arg;


    j = 1;
    for (i = 1; i < *argc; i ++)
	if (streq ((arg = argv [i]), "-help")) {
	    fputs (usage, stderr);
	    exit (0);
	} else if (streq (arg, "-debug")) {
	    debug = 1;
	} else if (streq (arg, "-printk")) {
	    printk = 1;
	} else if (streq (arg, "-summary")) {
	    summary = 1;
	} else if (streq (arg, "-plot")) {
	    doplot = 1;
	} else if (streq (arg, "+plot")) {
	    doplot = 0;
	} else if (streq (arg, "-table")) {
	    dotable = 1;
	} else if (streq (arg, "+table")) {
	    dotable = 0;
        } else if (streq (arg, "-renumber")) {
            renumber = 1;
	} else if (streq (arg, "-graphics")) {
	    if (++ i == *argc) {
		fputs (usage, stderr);
		return 1;
	    }
	    graphics = argv [i];
	} else
	    argv [j ++] = arg;

    argv [*argc = j] = NULL;
    return 0;
}


/************************************************************************
 * Function:	 main							*
 *									*
 * Description:	 Main is the driver function for the felt package.	*
 ************************************************************************/

int main (argc, argv)
    int   argc;
    char *argv [ ];
{
    char	*title;			/* title of problem		*/
    Element	*element;		/* array of element structures	*/
    Node	*node;			/* array of node structures	*/
    unsigned	 numelts,		/* number of element structures	*/
		 numnodes;		/* number of node structures	*/
    unsigned	 count;			/* number of affected dofs	*/
    Vector	 K,			/* global stiffness matrix	*/
		 Kcond;			/* condensed stiffness matrix	*/
    Vector	 F,			/* force vector			*/
		 Fcond,			/* condensed force vector	*/
		 d;			/* displacement vector		*/
    unsigned	 dofs [7];		/* degrees of freedom		*/
    int		 status;		/* return status		*/
    Reaction	 *R;			/* reaction force vector	*/
    unsigned	 numreactions;		/* the number of reactions	*/
    unsigned	 *diag;			/* diagonal address array	*/
    Vector	 M;			/* global mass matrix		*/
    Vector	 C;			/* global damping matrix	*/
    Matrix	 dtable;		/* time-displacement table	*/
    unsigned	 *old_numbers;		/* original node numbering	*/

        /*
         * Do everything to setup the problem
         */

    if (ParseCppOptions (&argc, argv)) {
	fputs (usage, stderr);
	exit (1);
    }

    if (ParseFeltOptions (&argc, argv)) {
	fputs (usage, stderr);
	exit (1);
    }

    if (argc > 2) {
	fputs (usage, stderr);
	exit (1);
    }

    if (ReadFeltFile (argc == 2 ? argv [1] : "-"))
	exit (1);

    element  = problem.elements;
    node     = problem.nodes;
    numelts  = problem.num_elements;
    numnodes = problem.num_nodes;
    title    = problem.title;


    if (numnodes == 0 || numelts == 0) 
       Fatal ("nothing to do");

	/*
	 * If debugging write the problem as we understand it to stdout
	 */

    if (debug) 
	WriteFeltFile ("-");

	/*
	 * Write a graphics file if necessary
	 */

    if (graphics != NULL) {
       status = WriteGraphicsFile (graphics, element, numelts);
       if (status)
          Fatal ("could not open graphics file for output");
    }

	/*
	 * find all the active DOFs in this problem	
	 */

    count = FindDOFS (element, numelts, dofs);

	/*
	 * renumber the nodes if desired
	 */

    old_numbers = NULL;

    if (renumber) 
	old_numbers = RenumberNodes (node, element, numnodes, numelts);

	/*
	 * switch on the problem type (transient or static)
	 */

    switch (problem.mode) {

       case Transient:
          status = CheckAnalysisParameters ( );

          if (status) 
             Fatal ("%d Errors found in analysis parameters.", status);

          status = AssembleDynamic (&K, &M, &C, element, numelts, numnodes, 
                                    count, dofs, &diag);
 
          if (status)
             Fatal ("%d fatal errors in stiffness and mass definitions",status);

          dtable = NewmarkIterate (K, M, C, node, numnodes, 
                                   count, dofs, diag, old_numbers);
          if (dtable == NullMatrix)
             Fatal ("fatal error in integration (probably a singularity).");

          if (dotable)
             WriteTransientTable (dtable, stdout);
    
          if (doplot)
             PlotTransientTable (dtable,analysis.duration,analysis.dt,stdout);

          break;
       
       case Static:

          K = AssembleStiffness(element,numelts,numnodes,count,dofs,
                                &diag,&status);

          if (status)

             Fatal ("%d Fatal errors in element stiffness definitions", status);

          if (printk)
             PrintCompactMatrix (K, diag, numnodes*count, stdout);

          F = ConstructForceVector (node, numnodes, count, dofs);
    
          ResolveConstraints (K, F, &Kcond, &Fcond, node, numnodes, 
                              count, dofs, diag);
     
          d = SolveForDisplacements (Kcond, Fcond, node, numnodes, 
                                     count, dofs, diag);
          if (d == NullVector)
             exit (1);
    
          status = EltStresses (element, numelts);

          if (status) 
             Fatal ("%d Fatal errors found computing element stresses", status);

          numreactions = SolveForReactions (K, d, node, numnodes, old_numbers,
                                            count, dofs, diag, &R);

          if (old_numbers != NULL)
             RestoreNodeNumbers (node, old_numbers, numnodes);

          WriteOutput (stdout, title, element, node, R, 
                       numelts, numnodes, numreactions, summary);

          break;

       case Modal:
          printf ("Modal analysis is not yet supported\n");
          break;
    }

    exit (0);
}
