/*
    This file is part of the FElt finite element analysis package.
    Copyright (C) 1993,1994 Jason I. Gobat and Darren C. Atkinson

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

# include <stdio.h>
# include <unistd.h>
# include <X11/Intrinsic.h>
# include <X11/StringDefs.h>
# include <X11/Xaw/Simple.h>
# include "Canvas.h" 
# include "Element.h"
# include "Node.h"
# include "Solution.h"
# include "allocate.h"
# include "fe.h"
# include "Tree.h"
# include "problem.h"
# include "globals.h"
# include "Drawing.h"
# include "vfe.h"
# include "procedures.h"
# include "error.h"

extern ElementDialog element_d;
extern NodeDialog    node_d;

/****************************************************************************
 * Functions:	BuildNodeArray and BuildElementArray
 *
 * Description:	These are two tree iterators that build the node and element
 *		arrays from the node and element trees
 ****************************************************************************/

static unsigned	array_count;

static int BuildNodeArray (item)
   Item		item;
{
   FigureAttributes	attr;
   char			buffer [10];
   Node			node;
   Drawn		drawn; 

   node = (Node) item;
   problem.nodes [++array_count] = node;

   if (array_count != node -> number) {
      node -> number = array_count;

      drawn = (Drawn) node -> aux;
      if (drawn -> label != NULL) {
	  sprintf (attr.text = buffer, " %d", node -> number);
	  DW_SetAttributes (drawing, drawn -> label, DW_FigureText, &attr);
      }
   }

   return 0;
}

static int BuildElementArray (item)
   Item		item;
{
   FigureAttributes	attr;
   char			buffer [10];
   Element		element;
   Drawn		drawn; 

   element = (Element) item;
   problem.elements [++array_count] = element;

   if (array_count != element -> number) {
      element -> number = array_count;

      drawn = (Drawn) element -> aux;
      if (drawn -> label != NULL) {
         ComputeCenter (element, &attr.x, &attr.y);
         sprintf (attr.text = buffer, "%d", element -> number);
         DW_SetAttributes (drawing, drawn -> label, DW_FigureText |
                                DW_FigureLocation, &attr);
      }
   }

   return 0;
}

/****************************************************************************
 *
 * Function:	SolveProblem
 *
 * Description:	Basically main from the Felt application ... pretty 
 *		straightforward
 *
 ***************************************************************************/

int SolveProblem ( )
{
    unsigned	 numnodes;		/* total number of nodes	*/
    unsigned	 numelts;		/* total number of elements	*/
    Node	*node;			/* the array of nodes		*/
    Element	*element;		/* array of elements		*/
    int		 i;			/* loop variable		*/
    unsigned	 count;			/* number of affected dofs	*/
    Vector	 K,			/* global stiffness matrix	*/
		 Kcond;			/* condensed stiffness matrix	*/
    Vector	 F,			/* force vector			*/
		 Fcond,			/* condensed force vector	*/
		 d;			/* displacement vector		*/
    unsigned	 dofs [7];		/* degrees of freedom		*/
    int		 status;		/* return status		*/
    Reaction	 *R;			/* reaction force vector	*/
    unsigned	 numreactions;		/* the number of reactions	*/
    unsigned	 *diag; 
    char	*temp_out;
    FILE	*output;
    unsigned	*old_numbers;
    Vector	M, C;
    Matrix	dtable;
    int		error_flag;

   	/*
	 * make sure the node and element numbering doesn't have any gaps
	 * and set-up the node and element arrays
	 */

    numnodes = CompactNodeNumbers ( );
    numelts  = CompactElementNumbers ( );
    node     = problem.nodes;
    element  = problem.elements;

    if (numnodes == 0 || numelts == 0) {
       error ("nothing to do!");
       return 1;
    }

    ClearNodes (node, numnodes);

	/*
	 * create a temporary file for output
	 */

    temp_out = tempnam (NULL,"velv");
    output = fopen (temp_out,"w");
    if (output == NULL) {
       error ("Could not open temporary file for output.");
       return 1;
    }

	/*	
	 * initialize all these so we won't have trouble destroying them
	 * if they never actually get allocated
	 */

    K = C = M = NullVector;
    F = Kcond = Fcond = d = NullVector;
    dtable = NullMatrix;
    diag = NULL;
    
	/*
	 * find the active DOFs and renumber the nodes if desired (common
	 * steps in all problem modes
	 */

    count = FindDOFS (element, numelts, dofs);

    if (solution -> renumber)
       old_numbers = RenumberNodes (node, element, numnodes, numelts);
    else
       old_numbers = NULL;

    error_flag = 0;

	/*
	 * switch on the problem mode depending on whether we not a
	 * static and transient analysis
	 */

    switch (problem.mode) {

    case Transient:
       status = CheckAnalysisParameters ( );
    
       if (status) {
          error ("%d errors found in analysis parameters.", status);
          error_flag = 1;
          break;
       }

       status = AssembleDynamic (&K, &M, &C, element, numelts, numnodes,
                                 count, dofs, &diag);
     
       if (status) {
          error ("%d errors in stiffness and mass definitions.", status);
          error_flag = 1;
          break;
       }

       dtable = NewmarkIterate (K, M, C, node, numnodes, count, 
                                dofs, diag, old_numbers);

       if (dtable == NullMatrix) {
          error ("could not perform integration - probably a singularity.");
          error_flag = 1;
          break;
       }
      
       if (solution -> table)
          WriteTransientTable (dtable, output);

       if (solution -> plot)
          PlotTransientTable (dtable, analysis.duration, analysis.dt, output);

       if (old_numbers != NULL)
          RestoreNodeNumbers (node, old_numbers, numnodes);
       
       break;

    case Static:

       K = AssembleStiffness (element,numelts,numnodes,
                              count,dofs,&diag,&status);

       if (status) {
   	  error ("%d fatal errors in element stiffness definitions.", status);
          error_flag = 1;
          break;
       }

       F = ConstructForceVector (node, numnodes, count, dofs);

       ResolveConstraints (K,F,&Kcond,&Fcond,node,numnodes,count,dofs,diag);

	/*
	 * the result will not have new space - it will get returned
	 * in the same space that Fcond used to occupy ...
	 */

       d = SolveForDisplacements (Kcond, Fcond, node, numnodes, 
                                  count, dofs, diag);

       if (d == NullVector) {
          error ("singluarity in final system of equations - cannot proceed.");
          error_flag = 1;
          break;
       }    

       numreactions = SolveForReactions (K,d,node,numnodes,old_numbers,
                                         count,dofs,diag,&R);

       status = EltStresses (element, numelts);
       if (status) {
          error ("%d errors found computing element stresses.",status);
          error_flag = 1;
          break;
        }   

       if (old_numbers != NULL)
          RestoreNodeNumbers (node, old_numbers, numnodes);

       WriteOutput (output, solution -> title, element, node, R,
                    numelts, numnodes, numreactions, solution -> summary);

       if (numreactions) {
   	  for (i = 1; i <= numreactions; i ++)
	     XtFree ((char *) R [i]);

	  XtFree ((char *) (R + 1));
       }

       break;

    case Modal:
	break;
    }

    if (dtable != NullMatrix) 
       DestroyMatrix (dtable);
    if (K != NullVector)
       DestroyVector (K);
    if (M != NullVector) 
       DestroyVector (M);
    if (C != NullVector) 
       DestroyVector (C);
    if (F != NullVector) 
       DestroyVector (F);
    if (Kcond != NullVector) 
       DestroyVector (Kcond);

    if (d == NullVector && Fcond != NullVector)
       DestroyVector (Fcond);
    else if (d != NullVector) 
       DestroyVector (d);

    if (diag) {
       ZeroOffset (diag);
       Deallocate (diag);
    }

    if (old_numbers) {
       ZeroOffset (old_numbers); 
       Deallocate (old_numbers);
    }

    fclose (output);
   
    if (error_flag)
       return 1;
 
    OutputDialogView (output_dialog, temp_out, 20, 80);

    unlink (temp_out);
    XtFree ((char *) temp_out);

    if (solution -> felt) 
       OutputWindowPopup ( );

    if (solution -> stress) 
       SetupStresses (False);
       
    if (solution -> structure)
       SetupStructure (False);

    return 0;
}

int CompactNodeNumbers ( )
{
    unsigned		numnodes;

    numnodes = TreeSize (problem.node_tree);
    if (numnodes == 0)
       return 0;

	/*
	 * build the array of _consecutive_ pointers to nodes by
	 * creating new space for it and then iterating on the tree
	 */

    if (problem.nodes) {
       ZeroOffset (problem.nodes); 
       Deallocate (problem.nodes);
    }
    problem.nodes = Allocate (Node, numnodes);
    UnitOffset (problem.nodes);

    DW_SetAutoRedraw (drawing, False);

    array_count = 0;
    TreeSetIterator (problem.node_tree, BuildNodeArray);
    TreeIterate (problem.node_tree);

    if (canvas -> node_numbers)
       DW_SetAutoRedraw (drawing, True);

    return problem.num_nodes = numnodes;
}          
    
int CompactElementNumbers ( )
{
    unsigned		numelts;

    numelts = TreeSize (problem.element_tree);
    if (numelts == 0)
       return 0;

    if (problem.elements) {
       ZeroOffset (problem.elements); 
       Deallocate (problem.elements);
    }

    problem.elements = Allocate (Element, numelts);
    UnitOffset (problem.elements);

    DW_SetAutoRedraw (drawing, False);

    array_count = 0;
    TreeSetIterator (problem.element_tree, BuildElementArray);
    TreeIterate (problem.element_tree);

    if (canvas -> element_numbers)
       DW_SetAutoRedraw (drawing, True);

    return problem.num_elements = numelts;
}

void SetupAndSolve ( )
{
    BufferErrors (True);
    SolveProblem ( );
    BufferErrors (False);
    NodeDialogDisplay (node_d, NodeDialogActive (node_d));
    ElementDialogDisplay (element_d, ElementDialogActive (element_d));
}

void SetupAnimate ( )
{
    int			status1, status2;
    unsigned		i,j;
    unsigned		count;
    unsigned		dofs [7];
    Vector		M, C, K;
    unsigned		*diag;
    unsigned		numnodes, numelts;
    Matrix		dtable;
    unsigned		*old_numbers;
    unsigned		*anodes;
    char		adofs [7];
    unsigned		anumnodes;
    unsigned		anumdofs;
    double		z_plane;
    Boolean		draw3d;

    if (problem.mode != Transient) {
       error ("problem mode must be transient for animation\n");
       return;
    }

    BufferErrors (True);

    numnodes = CompactNodeNumbers ( );
    if (numnodes == 0)
       return;

    numelts = CompactElementNumbers ( );
    if (numelts == 0)
       return;

    draw3d = False;
    z_plane = problem.elements[1] -> node[1] -> z;

    for (i = 1; i <= numelts ; i++) {
       for (j = 1 ; j <= problem.elements[i] -> definition -> numnodes ; j++) {
       
          if (problem.elements[i] -> node[j] -> z != z_plane) {
             draw3d = True;
             break;
          }
       }
       if (draw3d)
          break;
    }
 
    anumdofs = analysis.numdofs;
    for (i = 1 ; i <= analysis.numdofs ; i++)
       adofs [i] = analysis.dofs [i];

    anodes = analysis.nodes;
    analysis.nodes = Allocate (unsigned, numnodes);
    UnitOffset (analysis.nodes);

    anumnodes = analysis.numnodes;
    analysis.numnodes = numnodes;
    for (i = 1 ; i <= numnodes ; i++) 
       analysis.nodes [i] = problem.nodes [i] -> number;

    if (draw3d) {
       analysis.numdofs = 3;
       analysis.dofs [1] = Tx;
       analysis.dofs [2] = Ty;
       analysis.dofs [3] = Tz;
    }
    else {
       analysis.numdofs = 2;
       analysis.dofs [1] = Tx;
       analysis.dofs [2] = Ty;
    }


    if (solution -> renumber)
       old_numbers = RenumberNodes (problem.nodes, problem.elements, 
                                    numnodes, numelts);
    else
       old_numbers = NULL;

    K = M = C = dtable = NullMatrix;
    diag = NULL;
 
    count = FindDOFS (problem.elements, numelts, dofs);

    status1 = CheckAnalysisParameters ( );

    if (!status1)
       status2 = AssembleDynamic (&K, &M, &C, problem.elements, numelts,
                                  numnodes, count, dofs, &diag);

    if (status1 || status2) 
       error ("errors found - could not animate.");
    else {
       dtable = NewmarkIterate (K, M, C, problem.nodes, numnodes, count,
                                dofs, diag, old_numbers);

       if (draw3d)
          AnimateStructure3D (dtable, problem.nodes, problem.elements,
                              numnodes, numelts);
       else
          AnimateStructure (dtable, problem.nodes, problem.elements,
                            numnodes, numelts);

       if (old_numbers)
          RestoreNodeNumbers (problem.nodes, old_numbers, numnodes);
    }

    ZeroOffset (analysis.nodes);
    Deallocate (analysis.nodes);
    analysis.nodes = anodes;
    analysis.numnodes = anumnodes;
    analysis.numdofs = anumdofs;
    for (i = 1 ; i <= analysis.numdofs ; i++) 
       analysis.dofs [i] = adofs [i];

    if (old_numbers) {
       ZeroOffset (old_numbers);
       Deallocate (old_numbers);
    }
      
    if (diag) {
       ZeroOffset (diag);
       Deallocate (diag);
    }

    if (K != NullMatrix) 
       DestroyVector (K);
    if (M != NullMatrix) 
       DestroyVector (M);
    if (C != NullMatrix) 
       DestroyVector (C);
    if (dtable != NullMatrix) 
       DestroyMatrix (dtable);

    BufferErrors (False);
}

void SetupStresses (build_elt)
    Boolean	build_elt;
{
    unsigned	numelts;
    unsigned	nd,i;
    int		depth;
    Arg		arglist [1];

	/*
	 * technically someone on a 1-bit screen should never get here ...
	 * but just to be sure.
	 */

    XtSetArg (arglist[0], XtNdepth, &depth);
    XtGetValues (toplevel, arglist, 1);
    if (depth != 8) {
        error ("you must have an 8-bit deep screen to do stress plots");
        return;
    }

    if (build_elt) 
        numelts = CompactElementNumbers ( );

    if (numelts == 0) 
       return;

    nd = 0;
    for (i = 1; i <= numelts ; i++) {
         if (problem.elements[i] -> definition -> shape != Planar) {
            error ("cannot plot stresses for non-planar elements");
            return;
         }
         if (problem.elements[i] -> stress == NULL) {
            error ("could not get stresses for all elements");
            return;
         }
         if (4 > problem.elements[i] -> definition -> numstresses) {
            error ("invalid stress component for element %d",i);
            return;
         }
         nd += problem.elements[i] -> ninteg;
    }

    PlotStressField (problem.elements, numelts, nd);

    return;
}


void SetupStructure (build_elt) 
    Boolean	build_elt;
{
    unsigned	i,j;
    double	z_plane;
    Boolean	draw3d;
    unsigned	numelts;

    if (build_elt) 
       numelts = CompactElementNumbers ( );

    if (numelts == 0)
       return;

    draw3d = False;
    z_plane = problem.elements[1] -> node[1] -> z;

    for (i = 1; i <= numelts ; i++) {
       for (j = 1 ; j <= problem.elements[i] -> definition -> numnodes ; j++) {
       
          if (problem.elements[i] -> node[j] -> z != z_plane) {
             draw3d = True;
             break;
          }
       }
       if (draw3d)
          break;
    }
 
    if (draw3d)
       VisualizeStructure3D (problem.elements, numelts);
    else
       VisualizeStructure (problem.elements, numelts);
}

void OptimizeNumbering ( )
{
    char		buffer [10];
    FigureAttributes	attr;
    unsigned		numnodes, numelts;
    unsigned		i;
    Drawn 		drawn;
    Node		node;

    numnodes = CompactNodeNumbers ( );
    if (numnodes == 0)
       return;

    numelts = CompactElementNumbers ( );
    if (numelts == 0) 
       return;
  
    for (i = 1 ; i <= numnodes ; i++) 
       TreeDelete (problem.node_tree, problem.nodes [i]);

    (unsigned *) RenumberNodes (problem.nodes, problem.elements, 
                                numnodes, numelts);

    DW_SetAutoRedraw (drawing, False);
    for (i = 1 ; i <= numnodes ; i++) {
       node = problem.nodes [i];
       TreeInsert (problem.node_tree, node);

       drawn = (Drawn) (node -> aux);
       if (drawn -> label != NULL) {     
          sprintf (attr.text = buffer, " %d", node -> number); 
          DW_SetAttributes (drawing, drawn -> label, DW_FigureText, &attr);
       }
    }

    if (canvas -> node_numbers)
       DW_SetAutoRedraw (drawing, True);

    return;
}
