/*
    This file is part of the FElt finite element analysis package.
    Copyright (C) 1993,1994 Jason I. Gobat and Darren C. Atkinson

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/***************************************************************************
*
* File:		dynamic.c
*
* Description:	Contains code to implement dynamic analysis features of 
*		the finite element method.
*
* Notes:	The routines in this file are _closely_ tied in with
*		those in fe.c (the main algorithmic engine stuff).  The
*		only reason they're here instead of there is for 
*		simplicity.
*
* History:	v1.31 (or is this v2.0?) by Jason Gobat and Darren Atkinson	
*
***************************************************************************/

# include <stdio.h>
# include <math.h>
# include "fe.h"
# include "allocate.h"
# include "error.h"
# include "problem.h"

/****************************************************************************
*
* Function:	AssembleDynamic
*
* Parameters:	element		array of elements
*		numelts		total number of elements
*		numnodes	total number of nodes
*		count		number of DOFs per node
*		dofs		master index of affected DOFs
*		diag		diagonal addresses (to be set)
*
* Return value: the global stiffness matrix, K
*
* Calls:	CreateMatrix, ZeroMatrix
*
* Called by:	FELT driver application
*
* Global data:	None
*
* Description: 	See the description of AssembleStiffness () in fe.c.
*		This routine does the same thing except it includes
*		code to assemble the global mass matrix in addition
*		to the global stiffness matrix.  Having two separate
*		routines is basically a performance consideration 
*		(i.e., why do all the mass checks in the static case?)
*
****************************************************************************/

int AssembleDynamic (Kr, Mr, Cr, element, numelts, numnodes, count, dofs, diag)
   Vector	*Kr, *Mr, *Cr;
   Element	*element;
   unsigned	numelts,
		numnodes;
   unsigned	count;
   unsigned	*dofs;
   unsigned	**diag;
{
   Vector	M, K, C;
   unsigned	row,
		col,
		i,
		j,
		l,
		k,
		m;
   unsigned	size,
		ndofs,
		nodes;
   unsigned	base_row,
		base_col,
		affected_row_dof,
		affected_col_dof;
   unsigned	*ht,*dg;
   unsigned	address;
   double	mvalue;
   double	kvalue;
   int	 	err,
		err_count;

   err_count = 0;

	/*	
	 * first we make a pass over the elements to see how all the 
	 * stiffnesses fit together so we can set up our compact column
	 * storage sceme.  Inefficient as hell I concede.
	 */

   size = numnodes*count;

   ht = Allocate (unsigned, size);
   if (ht == NULL)
      Fatal ("allocation error setting up compact column heights");

   UnitOffset (ht);

   dg = Allocate (unsigned, size);
   if (dg == NULL)
      Fatal ("allocation error setting up compact column diagonal addresses");

   UnitOffset (dg);

   for (i = 1; i <= size ; i++) 
      ht[i] = dg[i] = 0;

   for (i = 1 ; i <= numelts ; i++) {
      err = EltSetup (element [i], analysis.mass_mode);

      if (err) {
         err_count += err;
         continue;
      } 
         
      ndofs = element[i] -> definition -> numdofs;
      nodes = element[i] -> definition -> numnodes;
      
      for (j = 1 ; j <= nodes ; j++) {
         if (element [i] -> node[j] == NULL) continue;
         base_row = (element[i] -> node[j] -> number - 1)*count + 1;

         for (k = 1 ; k <= nodes ; k++) {
            if (element [i] -> node[k] == NULL) continue;
            base_col = (element[i] -> node[k] -> number - 1)*count + 1;

            for (l = 1 ; l <= ndofs ; l++) {
               affected_row_dof = dofs[element[i] -> definition -> dofs[l]];
               row = base_row + affected_row_dof - 1;

               for (m = 1 ; m <= ndofs ; m++) {
                  affected_col_dof = dofs[element[i] -> definition -> dofs[m]];
                  col = base_col + affected_col_dof - 1;
                  kvalue =  MatrixData (element[i] -> K) [(j-1)*ndofs + l]
                                                         [(k-1)*ndofs + m]; 
                  mvalue =  MatrixData (element[i] -> M) [(j-1)*ndofs + l]
                                                         [(k-1)*ndofs + m]; 
                  if ((kvalue != 0.0 || mvalue != 0.0) && row <= col) { 
                     if (col-(row-1) > ht [col])
                        ht [col] = col - (row - 1);
                  }
               }
            }
         }
      }
   } /* end first loop over elements */

   if (err_count) 
      return err_count;

	/*
	 * setup the diagonal address array and figure out how big
	 * we need to make the compact column vector
	 */

   dg[1] = 1;
   size = 1;
   if (ht [1] == 0)
      ht [1] = 1;

   for (i = 2 ; i <= numnodes*count ; i++) {
      if (ht[i] == 0)
         ht[i] = 1;

      size += ht[i];
      dg [i] = ht [i] + dg [i-1];
   }

   ZeroOffset (ht);
   Deallocate (ht);

   K = CreateVector (size);
   if (K == NullMatrix)
      Fatal ("could not allocate memory for global stiffness matrix");

   M = CreateVector (size);
   if (M == NullMatrix)
      Fatal ("could not allocate memory for global mass matrix");

   C = CreateVector (size);
   if (C == NullMatrix)
      Fatal ("could not allocate memory for global damping matrix");

   ZeroMatrix (K);
   ZeroMatrix (M);
   ZeroMatrix (C);

	/*
	 * now we make just about the identical passes over the elements,
	 * the result of this pass however will be that we actually
	 * start sticking stuff into the vector which is the compact
	 * column representation of the global stiffness matrix
	 */

   for (i = 1 ; i <= numelts ; i++) {
      ndofs = element[i] -> definition -> numdofs;
      nodes = element[i] -> definition -> numnodes;
      
      for (j = 1 ; j <= nodes ; j++) {
         if (element [i] -> node[j] == NULL) continue;
         base_row = (element[i] -> node[j] -> number - 1)*count + 1;

         for (k = 1 ; k <= nodes ; k++) {
            if (element [i] -> node[k] == NULL) continue;
            base_col = (element[i] -> node[k] -> number - 1)*count + 1;

            for (l = 1 ; l <= ndofs ; l++) {
               affected_row_dof = dofs[element[i] -> definition -> dofs[l]];
               row = base_row + affected_row_dof - 1;

               for (m = 1 ; m <= ndofs ; m++) {
                  affected_col_dof = dofs[element[i] -> definition -> dofs[m]];
                  col = base_col + affected_col_dof - 1;
                  kvalue =  MatrixData (element[i] -> K) [(j-1)*ndofs + l]
                                                         [(k-1)*ndofs + m]; 
                  mvalue =  MatrixData (element[i] -> M) [(j-1)*ndofs + l]
                                                         [(k-1)*ndofs + m]; 

                  if (row <= col) {
                     address = ConvertRowColumn (row, col, dg);
                     if (address) {
                        VectorData (K) [address] += kvalue;
                        VectorData (M) [address] += mvalue;
                        VectorData (C) [address] += 
                                   element[i] -> material -> Rk * kvalue +
                                   element[i] -> material -> Rm * mvalue;
                     }
                  }
               }
            }
         }
      }

      if (!element[i] -> definition -> retainK) {
         DestroyMatrix (element[i] -> K);
         element[i] -> K = NullMatrix;
      }

      DestroyMatrix (element[i] -> M);

   } /* end second loop over elements */

	/*
	 * set some things up for the return
	 */

   *diag = dg;

   *Kr = K;
   *Mr = M;
   *Cr = C;

   return 0;
}

/****************************************************************************
*
* Function:	ConstructTransientForce
*
* Parameters:	t		time t to evaluate F(t) for
*		node		array of nodes
*		numnodes 	total number of nodes in problem
*		count		number of DOFs per node
*		dofs		master index of affected DOFs
*
* Return value:	
*
* Calls:	
*
* Called by:	
*
* Global data:	None
*
* Description:	 
*
****************************************************************************/

void ConstructTransientForce (t,node,numnodes,count,dofs,Fadj,F)
   double	t;
   Node		*node;
   unsigned	numnodes,
		count;
   unsigned	*dofs;
   Vector	Fadj;
   Vector	F;
{
   unsigned	i,j,
		base_dof;
   unsigned	size;
   double	force;
   double	factor;

   size = numnodes*count;

   for (i = 1 ; i <= size ; i++) 
      VectorData (F) [i] = VectorData (Fadj) [i];

   for (i = 1 ; i <= numnodes ; i++) {

      base_dof = count*(node[i] -> number - 1) + 1;

      for (j = 1 ; j <= 6 ; j++) {
         if (dofs[j]) {
            force = 0.0;
            if (node[i] -> force != NULL) {
               if (node[i] -> force -> expr [j] != NULL) {
                  factor = EvalCode (node[i] -> force -> expr [j], t);
                  force += factor;
               }
               else if (node[i] -> force -> force[j]) 
                  force += node[i] -> force -> force[j];
            }
            if (node[i] -> eq_force != NULL) {
               if (node[i] -> eq_force[j])
                  force += node[i] -> eq_force[j];
            }
            VectorData (F) [base_dof + dofs[j] - 1] += force;
         }
      }
   }

   return;
}

/*****************************************************************************
*
* Function:	NewmarkIterate
*
* Parameters:	K		the compact column stiffness matrix
*		M		the compact column mass matrix
*		C		the compact column damping matrix
*		node		array of problem nodes
*		numnodes	number of nodes in the problem
*		count		number of DOFs per node of problem
*		dofs		array of problem DOFs
*		diag		diagonal addresses in compact columns
*
* Return value:	None
*
* Calls: 	CreateVector, CreateMatrix, MultiplyCompactColumn,
*		BuildInitialConditions, ResolveInitialConstraints,
*		CompactForwardBack, CroutFactorization, ConvertRowColumn,
*		ConstructTransientForce, DestroyVector
*
* Called by:	FElt driver application
*
* Global data: 	None
*
* Description:  Solves the discrete equation of motion, Ma + Cv + Kd = F
*		for the length of a model using Newmark's method
*		with the Hilber-Hughes-Taylor alpha correction for
*		improved accuracy with numerical damping.
*
*		The first important numerical thing that we do is
*		to solve for the initial acceleration vector:
*		Ma(0) = F(0) - Kd(0) - Cv(0). From there we can begin
*		the iterations - the iterations proceed by solving
*		for a(i+1) implicity (not d(i+1)!) and then using this
*		information with Newmark's update equations to get
*		d(i+1) and v(i+1)
*
******************************************************************************/

Matrix NewmarkIterate (K, M, C, node, numnodes, count, dofs, diag, old_numbers)
   Vector	K;
   Vector	M;
   Vector	C;
   Node		*node;
   unsigned	numnodes;
   unsigned	count;
   unsigned	*dofs;
   unsigned	*diag;
   unsigned	*old_numbers;
{
   unsigned	i,j;
   Matrix	dtable;
   Vector	d;
   Vector	a;
   Vector	v;
   Vector	F;
   Vector	Kp;
   Vector	Mt;
   Vector	Fadj;
   unsigned	*constraint_mask;
   double	vpred, dpred, value;
   unsigned	size;
   double	c1,c2, c3, c4, c5;
   unsigned	step;
   unsigned	nsteps;
   int		address;
   unsigned	*node_map;
   int		build_a0; 

	/*
	 * a few constants that we will need
	 */

   size = numnodes*count;
   c1 = (1.0 - 2.0*analysis.beta) * (analysis.dt*analysis.dt)/2.0;
   c2 = analysis.dt * (1.0 - analysis.gamma);
   c3 = analysis.dt * analysis.dt * analysis.beta;
   c4 = analysis.dt * analysis.gamma;
   c5 = (1.0 + analysis.alpha);

	/*
	 * create vectors to hold the conditions at timesteps i and i+1
	 */

   d  = CreateVector (size);
   a  = CreateVector (size);
   v  = CreateVector (size);
   F  = CreateVector (size);   

	/*
	 * create the table of nodal time displacements
	 */

   nsteps = analysis.duration / analysis.dt + 1.0;
   dtable = CreateMatrix (nsteps, analysis.numnodes*analysis.numdofs);

	/*
	 * if some nodes got renumbered then we need to make a map
	 * that points from old numbering into the currently numbered
	 * vectors
	 */

   node_map = Allocate (unsigned, analysis.numnodes);
   UnitOffset (node_map);
   for (i = 1 ; i <= analysis.numnodes ; i++) {
      if (old_numbers != NULL) {
         for (j = 1 ; j <= numnodes ; j++) {
            if (old_numbers [j] == analysis.nodes [i]) {
               node_map [i] = node [j] -> number; 
               break;
            }
         } 
      } 
      else
         node_map [i] = analysis.nodes [i];
   }

	
	/*
	 * Resolve the _boundary_ conditions, i.e., modify the K'
	 * matrix such that fixed displacements are properly taken
	 * care of.  Instead of a condensed force vector what we'll
	 * fill in is the adjustment vector - this is what we need to
	 * add on to the transient force vector at each time step due
	 * to any displacement boundary conditions.
	 */

   Kp = AssignMatrix (NullMatrix, K);

   Fadj = CreateVector (size);
   constraint_mask = Allocate (unsigned, size); UnitOffset (constraint_mask);
   ResolveInitialConstraints (Kp, Fadj, constraint_mask, node, numnodes, 
                              count, dofs, diag);

	/*
	 * create the K' matrix
	 */

   for (i = 1 ; i <= VectorSize (K) ; i++)
      VectorData (Kp) [i] = VectorData (M) [i] + 
                            VectorData (C) [i]*c4*c5 + 
                            VectorData (Kp) [i]*c3*c5;

	/* 
	 * build the initial displacement and velocity vectors from the
	 * initial conditions	
 	 */

   build_a0 = BuildInitialConditions (node, numnodes, count, dofs, d, v, a);

	/*
	 * build the F(0) vector, we only need this to get a(0),
	 * after this, we really will use F as F(i+1)
	 */

   ConstructTransientForce (0.0, node, numnodes, count, dofs, Fadj, F);

	/*
	 * solve for the initial acceleration vector.  First we factorize
	 * the mass matrix, then we form the right hand side vector
	 * as F(0) - Kd(0) - Cv(0) and solve the system to get a(0)
	 */

   if (build_a0) {
      Mt = AssignMatrix (NullMatrix, M);
      if (CroutFactorization (Mt, diag, size)) {
         error ("singular M matrix in Newmark integration - could not proceed");
         return NullMatrix;
      }

      MultiplyCompactColumn (K, d, diag, a);
      SubtractMatrices (a, F, a);
      MultiplyCompactColumn (C, v, diag, F);
      SubtractMatrices (a, a, F);

      if (CompactForwardBack (Mt, a, diag, size)) {
         error ("singular M matrix in Newmark integration - could not proceed");
         return NullMatrix;
      }
   }
   else
      Mt = NullMatrix;

	/*
	 * perform Crout factorization on the K' matrix ... we'll only
	 * need to do this once
	 */

   if (CroutFactorization (Kp, diag, size)) {
      error ("singular K' matrix in Newmark integration - could not proceed");
      return NullMatrix;
   }

	/*
	 * Copy the initial displacement vector into the table.
	 * This is basically a copy of the code at the end of the loop.
	 */

   for (i = 1 ; i <= analysis.numnodes ; i++) {
      for (j = 1 ; j <= analysis.numdofs ; j++) {
         MatrixData (dtable) [1][(i-1)*analysis.numdofs + j] = 
           VectorData (d)[(analysis.nodes[i] - 1)*count + dofs [analysis.dofs[j]]];
      }
   }

	/*
	 * iterate over every time step.  Fill up dtable with
	 * the results
	 */

   for (step = 2 ; step <= nsteps ; step++) {
         
	/*
	 * setup F'(i+1).  First find F(i+1) = F(t + dt), then
	 * adjust it.
	 */

      ConstructTransientForce ((step - 1.0)*analysis.dt, node, numnodes, 
                               count, dofs, Fadj, F);      
      
	/*
	 * now we multiply Kd and Cv and subtract from F, but
	 * we'll just blob it all into one big step
	 */

      for (i = 1 ; i <= size ; i++) {
         if (!constraint_mask [i]) {

            value = 0.0;
            for (j = 1 ; j <= size ; j++) {
               address = ConvertRowColumn (i, j, diag);
               if (address) {
                  dpred = VectorData (d) [j] + 
                          analysis.dt*c5*VectorData (v) [j] +
                          c1*c5*VectorData (a) [j];
                  vpred = VectorData (v) [j] + c2*c5*VectorData (a) [j];

                  value += VectorData (C) [address]*vpred +
                           VectorData (K) [address]*dpred;
               }
            }
            VectorData (F) [i] -= value;
         }
      }

	/*
	 * solve for K'a(i+1) = F'(i+1) ... the result will go into F 
	 */

      CompactForwardBack (Kp, F, diag, size);
 
	/*
	 * from here we'll solve for d(i+1) and v(i+1)
	 */
      
      for (i = 1 ; i <= size ; i++) {
         VectorData (d) [i] += analysis.dt*VectorData (v)[i] +
                               c1*VectorData (a)[i] + c3*VectorData (F)[i]; 
         VectorData (v) [i] += c2*VectorData (a)[i] + c4*VectorData (F)[i];
         VectorData (a) [i] = VectorData (F) [i];
      }

	/*
	 * copy the relevant parts of the displacement vector
	 * into the displacement table
	 */

      for (i = 1 ; i <= analysis.numnodes ; i++) {
         for (j = 1 ; j <= analysis.numdofs ; j++) {
            MatrixData (dtable) [step][(i-1)*analysis.numdofs + j] = 
              VectorData (d)[(node_map [i] - 1)*count + 
                             dofs [analysis.dofs[j]]];
         }
      }
   }    

	/*
	 * clean up ... 
	 */

   DestroyVector (F);
   DestroyVector (a);
   DestroyVector (d);
   DestroyVector (v);
   DestroyVector (Kp);
   if (Mt)
      DestroyVector (Mt);

   ZeroOffset (constraint_mask); Deallocate (constraint_mask);
   ZeroOffset (node_map); Deallocate (node_map);

   return dtable;
}

/****************************************************************************
*
* Function:	BuildInitialConditions
*
* Parameters:	node		array of nodes
*		numnodes	total number of nodes in problem
*		count		number of DOFs per node
*		dofs		master index of affected DOFs
*		d		initial displacement vector
*		v		initial velocity vector
*
* Return value:	code to indicate whether a(0) was built or not
*
* Calls:	None
*
* Called by:	NewmarkIterate
*
* Global data: 	None
*
* Description:  Fills in the displacement and velocity vectors at time
*		t = 0 given the nodal constraint conditions.
*		
****************************************************************************/

int BuildInitialConditions (node, numnodes, count, dofs, d, v, a)
   Node		*node;
   unsigned	numnodes;
   unsigned	count;
   unsigned	*dofs;
   Vector	d, v, a;
{
   unsigned	i,j,
		base_dof;
   unsigned	size;
   int		build_a0;

   size = numnodes*count;

   for (i = 1 ; i <= size ; i++) {
      VectorData (d) [i] = 0.0;
      VectorData (v) [i] = 0.0;
      VectorData (a) [i] = 0.0;
   } 

   build_a0 = 1;

   for (i = 1 ; i <= numnodes ; i++) {

      base_dof = count*(node[i] -> number - 1);

      for (j = 1 ; j <= 6 ; j++) {
         if (dofs[j]) {
            VectorData (d) [base_dof+dofs[j]] = node[i] -> constraint -> dx[j];
            if (j <= 3) {
               VectorData (v) [base_dof+dofs[j]] = node[i] -> constraint -> vx[j];
               if (node [i] -> constraint -> ax[j] != UnspecifiedValue) {
                  build_a0 = 0;
                  VectorData (a) [base_dof+dofs[j]] = 
                                    node[i] -> constraint -> ax[j];
               }
            }
            else {
               VectorData (v) [base_dof+dofs[j]] = 0.0;
               VectorData (a) [base_dof+dofs[j]] = 0.0;
            }
         }
      }
   }

   return build_a0;
}

/****************************************************************************
*
* Function:	ResolveInitialConstraints
*
* Parameters:	Kcond		on return, the condensed stiffness matrix
*		Fadj		on return, the adjusting force vector 
*		node		array of nodes
*		constraint_mask mask array of constrained DOFs 
*		numnodes	total number of nodes in problem
*		count		number of DOFs per node
*		diag		diagonal addresses
*		dofs		master index of affected DOFs
*
* Return value:	None
*
* Calls:	ZeroCompactRowCol, AdjustForceVector
*
* Called by:	NewmarkIterate
*
* Global data: 	None
*
* Description:  Basically the same as ResolveConstraints () for the
*		static case, but here we are handling the matrices
*		differently, etc.  We'll also build a mask of which
*		dofs are active and which are constrained to make sure
*		we can do this check very quickly during the timestepping
*		
****************************************************************************/

void ResolveInitialConstraints (Kcond,Fadj,constraint_mask,node,numnodes,
                                count,dofs,diag)
   Vector	Kcond;
   Vector	Fadj; 
   unsigned	*constraint_mask;
   Node		*node;
   unsigned	numnodes,
		count;
   unsigned	*dofs;
   unsigned	*diag;
{
   unsigned	i,j,
		affected_dof,
		base_dof;

   ZeroMatrix (Fadj);

   for (i = 1 ; i <= numnodes ; i++) {
     
      base_dof = count*(node[i] -> number - 1) + 1;

      for (j = 1 ; j <= 6 ; j++) {
         if (dofs[j]) {
            affected_dof = base_dof + dofs[j] - 1;
            constraint_mask [affected_dof] = 0;
            if (node[i] -> constraint -> constraint[j]) {

               if (node[i] -> constraint -> dx[j] == 0.0 ||
                   node[i] -> constraint -> constraint[j] == 'h') {
                  Kcond = ZeroCompactRowCol (Kcond,affected_dof,
                                             diag,VectorSize(Fadj)); 
                  constraint_mask [affected_dof] = 1;
               }
               else {
                  Fadj = AdjustForceVector (Fadj, Kcond, affected_dof,
                                             node[i] -> constraint -> dx[j],
                                             numnodes*count, diag);
                  Kcond = ZeroCompactRowCol (Kcond,affected_dof,
                                             diag,VectorSize(Fadj)); 
                  constraint_mask [affected_dof] = 1;
                  VectorData (Fadj) [affected_dof] = 
                                             node[i] -> constraint -> dx[j];
               }
            }
         }
      }
   }

   return;
}

static char symbols[] = {0,'x','o','=','*','?','#','v','$','&','@'};
static char *labels [] = {"","Tx","Ty","Tz","Rx","Ry","Rz"};

/******************************************************************************
*
* Function:	WriteTransientTable
*
* Parameters:	dtable		matrix of nodal displacements with time
*		fp		file pointer to output table to
*
* Return value:	None
*
* Calls:	None
*
* Called by:	FElt driver application
*
* Global data: 	None
*
* Description:	Pretty simple really. The only slightly tricky part is that
*		we don't want to print a whole bunch of DOFs all in one
*		table ... so we'll break it up into 4 DOFs per table and
*		then print as many of them as we need
*	
******************************************************************************/

void WriteTransientTable (dtable, fp)
   Matrix	dtable;
   FILE		*fp;
{
   unsigned	i,j,k,m,n;
   unsigned	table;
   unsigned	node, dof;
   unsigned	start_dof;
   unsigned	number;
   unsigned	ntables;

   ntables = (unsigned) (analysis.numnodes * analysis.numdofs / 4);
   if ((analysis.numdofs * analysis.numnodes) % 4 != 0)
      ntables++;

	/*
	 * for each table, print the appropriate headers and then
	 * the displacement data for those same DOFs
	 */

   node = 1;
   dof = 1;
   for (table = 1 ; table <= ntables ; table++) {
      start_dof = dof;
      fprintf (fp,"\n------------------------------------------------------------------\n");
      fprintf (fp,"       time");
      number = 1;
      for (m = node ; m <= analysis.numnodes && number <= 4 ; m++) {
         for (n = dof ; n <= analysis.numdofs && number <= 4 ; n++) {
            fprintf (fp,"        %s(%d)", labels[(int) analysis.dofs[n]],
                     analysis.nodes[m]);
            
            number++; 
         }
         dof = 1;
      }
      fprintf (fp,"\n------------------------------------------------------------------\n");

	/*
	 * output the displacement at this time step for each DOF 
	 *in this table
	 */

      for (i = 1 ; i <= MatrixRows (dtable) ; i++) {
         fprintf (fp,"%11.5g",(i-1)*analysis.dt);
         number = 1;
         dof = start_dof;
         for (j = node ; j <= analysis.numnodes && number <= 4  ; j++) {
            for (k = dof ; k <= analysis.numdofs && number <= 4 ; k++) {
               fprintf (fp, "  %11.5g",
                        MatrixData (dtable)[i][(j-1)*analysis.numdofs + k]);
               number++;
            }
            dof = 1;
         }
         fprintf (fp,"\n");
      }
    
      node = m - 1;
      dof = n; 
   }
   fprintf (fp,"\n");

   return;
}

/******************************************************************************
*
* Function:	PlotTransientTable
*
* Parameters:	dtable		matrix of nodal displacements with time
*		length		length of simulation in time
*		dt		time step of simulation
*		fp		file pointer to output table to
*
* Return value:	None
*
* Calls:	None	
*
* Called by:	FElt driver application
*
* Global data: 	None
*
* Description:	Again this should be pretty simple but we can only plot 10 
*		DOFs per graph (even this is pushing cluttered).  So we plot
*		multiple graphs if we have to.
*	
******************************************************************************/

void PlotTransientTable (dtable, length, dt, fp)
   Matrix	dtable;
   double	length;
   double	dt;
   FILE		*fp;
{
   unsigned	i,j,k;
   unsigned	m,n;
   double	data;
   double	max;
   double	min;
   double	dscale;
   char		buffer [80];
   char		buffer1 [80];
   char		buffer2 [80];
   unsigned	graph, ngraphs;
   unsigned	start_dof, number;
   unsigned	dof, node;
   int		position;

   ngraphs = (unsigned) (analysis.numnodes * analysis.numdofs / 10);
   if ((analysis.numdofs * analysis.numnodes) % 10 != 0)
      ngraphs++;

   node = 1;
   dof = 1;
   for (graph = 1 ; graph <= ngraphs ; graph++) {

	/*
	 * find the max and the min for this graph, then compute the scale
	 */

      min = max = MatrixData (dtable) [1][(node-1)*analysis.numdofs + dof]; 
      for (i = 1 ; i <= MatrixRows (dtable) ; i++) {
         start_dof = dof;
         number = 1;

         for (m = node ; m <= analysis.numnodes && number <= 10 ; m++) {
            for (n = dof ; n <= analysis.numdofs && number <= 10 ; n++) {
        
               data = MatrixData (dtable)[i][(m-1)*analysis.numdofs + n];

               if (data < min)
                  min = data;
               else if (data > max)
                  max = data;

               number++;
            }
            dof = 1;
         }
      }

      dscale = 67.0/(max - min);
   
      fprintf (fp,"\n");
      fprintf (fp,"           %-11.5g             %11.5g                      %11.5g\n",min,(min+max)/2.0,max);
      fprintf (fp,"           +------------------------------+-----------------------------------+\n");

	/*
	 * for each line (time step) form a line of the plot by placing
	 * each DOF in the appropriate place in the line buffer.  If that
	 * space is occupied we try the line above it and then the line
	 * below it.  If all available spaces are filled, we skip this point.
	 */

      for (i = 1; i <= MatrixRows (dtable) ; i++) {
         number = 1;
         dof = start_dof;
         sprintf (buffer1,"|                                                                   ");
         sprintf (buffer2,"|                                                                   ");
         sprintf (buffer,"|                                                                   ");

         for (j = node ; j <= analysis.numnodes && number <= 10  ; j++) {
            for (k = dof ; k <= analysis.numdofs && number <= 10 ; k++) {
               data = MatrixData (dtable)[i][(j-1)*analysis.numdofs + k];
               position = (int) ((data - min)*dscale);
               if (buffer [position] == ' ' || buffer [position] == '|')
                  buffer [position] = (char) symbols [number];
               else {
                  if (buffer1 [position] == ' ' || buffer1 [position] == '|')
                     buffer1 [position] = (char) symbols [number];
                  else if (buffer2[position] == ' ' || buffer2[position] == '|')
                     buffer2 [position] = (char) symbols [number];
               }

               number++;
            }
            dof = 1;
         }
         fprintf (fp,"           %s\n",buffer1);
         fprintf (fp,"%10.5g %67s\n",(i-1)*dt,buffer);
         fprintf (fp,"           %s\n",buffer2);          
      }
      fprintf (fp, "\n");

	/*
	 * print the legend for this graph
	 */

      number = 1;
      dof = start_dof;
      for (i = node ; i <= analysis.numnodes && number <= 10 ; i++) {
         for (j = dof ; j <= analysis.numdofs && number <= 10 ; j++) {
            fprintf (fp,"         %c %c %c    %s(%d)\n", symbols [number],
                     symbols [number], symbols [number],
                     labels[(int) analysis.dofs[j]], analysis.nodes[i]);

            number++;
         }
         dof = 1;
      }
      node = m - 1;
      dof = n; 
   }
   fprintf (fp,"\n");

   return;
}

/***************************************************************************
*
* Function:	CheckAnalysisParameters
*
* Description:	Verifies that everything in the analysis parameters
*		section is set (or at least the minimum number of
*		things that we need).
*
***************************************************************************/

int CheckAnalysisParameters ( )
{
   unsigned	count;

   count = 0;

   if (analysis.mass_mode == 0) {
      error ("mass-mode must be defined for transient analysis");
      count++;
   }

   if (analysis.numnodes == 0) {
      error ("need to specify a node list for transient analysis");
      count++;
   }

   if (analysis.numdofs == 0) {
      error ("need to specify a list of DOFs for transient analysis");
      count++;
   }

   if (analysis.duration <= 0) {
      error ("duration needs to be greater than zero for transient analysis");
      count++;
   }

   if (analysis.dt <= 0) {
      error ("time step needs to be greater than zero for transient analysis");
      count++;
   }

   return count;
}
