/*
    This file is part of the FElt finite element analysis package.
    Copyright (C) 1993,1994 Jason I. Gobat and Darren C. Atkinson

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/************************************************************************
 * File:	file.c							*
 *									*
 * Description:	This file contains the public and private function	*
 *		definitions for writing FElt files and objects.		*
 ************************************************************************/

# include <stdio.h>
# include <ctype.h>
# include <string.h>
# include "fe.h"
# include "error.h"
# include "problem.h"

# define ELEMENTS
# include "element.h"

# define cfprintf(fp,f,x) if (x) fprintf (fp, f, x)

static char   *mark_flag;
static FILE   *fp;
static Node    prev_node;
static Element prev_element;


/************************************************************************
 * Function:	ConstraintSymbol					*
 *									*
 * Description:	Format a constraint into a proper string.		*
 ************************************************************************/

static char *ConstraintSymbol (constraint, dof)
    Constraint constraint;
    DOF	       dof;
{
    static char buffer [32];


    if (constraint -> constraint [dof] == 0)
	return "u";

    if (constraint -> constraint [dof] == 'h')
	return "h";

    if (constraint -> dx [dof] == 0)
	return "c";

    sprintf (buffer, "%g", constraint -> dx [dof]);
    return buffer;
}


/************************************************************************
 * Function:	Quote							*
 *									*
 * Description:	Quotes a string if necessary.				*
 ************************************************************************/

static char *Quote (s)
    char *s;
{
    char	c;
    char       *ptr;
    static char buffer [256];


    for (ptr = s; (c = *ptr); ptr ++)
	if (!isalpha (c) && !isdigit (c) && c != '_') {
	    sprintf (buffer, "\"%s\"", s);
	    return buffer;
	}

    return s;
}


/************************************************************************
 * Function:	WriteNode						*
 *									*
 * Description:	Writes a node to the specified stream.			*
 ************************************************************************/

static void WriteNode (node)
    Node node;
{
    fprintf (fp, "%-2u x=%g ", node -> number, node -> x);
    fprintf (fp, "y=%g z=%g", node -> y, node -> z);

    if (!prev_node || node -> constraint != prev_node -> constraint)
	fprintf (fp, " constraint=%s", Quote (node -> constraint -> name));

    if (node -> force)
	fprintf (fp, " force=%s", Quote (node -> force -> name));

    fprintf (fp, "\n");
}


/************************************************************************
 * Function:	WriteElement						*
 *									*
 * Description:	Writes an element to the specified stream.		*
 ************************************************************************/

static void WriteElement (element)
    Element element;
{
    unsigned i;
    unsigned numnodes;


    fprintf (fp, "%-2u nodes=[", element -> number);

    numnodes = element -> definition -> numnodes;
    for (i = 1; i <= numnodes; i ++) {
	if (element -> node [i])
	    fprintf (fp, "%u", element -> node [i] -> number);
	else
	    fprintf (fp, "0");
	fprintf (fp, i != numnodes ? "," : "]");
    }

    if (!prev_element || element -> material != prev_element -> material)
	fprintf (fp, " material=%s", Quote (element -> material -> name));

    for (i = 1; i <= element -> numdistributed; i ++) {
	fprintf (fp, i == 1 ? " load=" : " ");
	fprintf (fp, "%s", Quote (element -> distributed [i] -> name));
    }

    fprintf (fp, "\n");
}


/************************************************************************
 * Function:	WriteMaterial						*
 *									*
 * Description:	Writes a material to the specified stream if the aux	*
 *		pointer matches the mark flag.  The pointer is then	*
 *		cleared.						*
 ************************************************************************/

static int WriteMaterial (item)
    Item item;
{
    Material material;


    material = (Material) item;

    if (material -> aux == mark_flag) {
	fprintf (fp, "%s", Quote (material -> name));

	cfprintf (fp, " E=%g",     material -> E);
	cfprintf (fp, " A=%g",     material -> A);
	cfprintf (fp, " Ix=%g",    material -> Ix);
	cfprintf (fp, " Iy=%g",    material -> Iy);
	cfprintf (fp, " Iz=%g",    material -> Iz);
	cfprintf (fp, " J=%g",     material -> J);
	cfprintf (fp, " G=%g",     material -> G);
	cfprintf (fp, " nu=%g",    material -> nu);
	cfprintf (fp, " t=%g",     material -> t);
	cfprintf (fp, " rho=%g",   material -> rho);
	cfprintf (fp, " kappa=%g", material -> kappa);
	cfprintf (fp, " Rk=%g",    material -> Rk);
	cfprintf (fp, " Rm=%g",    material -> Rm);

	fprintf (fp, "\n");
    }

    material -> aux = NULL;
    return 0;
}


/************************************************************************
 * Function:	WriteLoad						*
 *									*
 * Description:	Writes a distributed load to the specified stream if	*
 *		the aux pointer matches the mark flag.  The pointer is	*
 *		then cleared.						*
 ************************************************************************/

static int WriteLoad (item)
    Item item;
{
    unsigned	 i;
    Distributed  load;
    static char *direction_names [ ] = {"", "LocalX", "LocalY", "LocalZ",
					"GlobalX", "GlobalY", "GlobalZ",
					"parallel", "perpendicular"};


    load = (Distributed) item;

    if (load -> aux == mark_flag) {
	fprintf (fp, "%s", Quote (load -> name));
	fprintf (fp, " direction=%s", direction_names [load -> direction]);

	for (i = 1; i <= load -> nvalues; i ++) {
	    fprintf (fp, i == 1 ? " values=" : " ");
	    fprintf (fp, "(%u,", load -> value [i].node);
	    fprintf (fp, "%g)", load -> value [i].magnitude);
	}

	fprintf (fp, "\n");
    }

    load -> aux = NULL;
    return 0;
}


/************************************************************************
 * Function:	WriteConstraint						*
 *									*
 * Description:	Writes a constraint to the specified stream if the aux	*
 *		pointer matches the mark flag.  The pointer is then	*
 *		cleared.						*
 ************************************************************************/

static int WriteConstraint (item)
    Item item;
{
    Constraint constraint;


    constraint = (Constraint) item;

    if (constraint -> aux == mark_flag) {
	fprintf (fp, "%s", Quote (constraint -> name));

	fprintf (fp, " Tx=%s", ConstraintSymbol (constraint, Tx));
	fprintf (fp, " Ty=%s", ConstraintSymbol (constraint, Ty));
	fprintf (fp, " Tz=%s", ConstraintSymbol (constraint, Tz));
	fprintf (fp, " Rx=%s", ConstraintSymbol (constraint, Rx));
	fprintf (fp, " Ry=%s", ConstraintSymbol (constraint, Ry));
	fprintf (fp, " Rz=%s", ConstraintSymbol (constraint, Rz));

	cfprintf (fp, " Vx=%g", constraint -> vx [Tx]);
	cfprintf (fp, " Vy=%g", constraint -> vx [Ty]);
	cfprintf (fp, " Vz=%g", constraint -> vx [Tz]);

	if (constraint -> ax [Tx] != UnspecifiedValue)
	    fprintf (fp, " Ax=%g", constraint -> ax [Tx]);

	if (constraint -> ax [Ty] != UnspecifiedValue)
	    fprintf (fp, " Ay=%g", constraint -> ax [Ty]);

	if (constraint -> ax [Tz] != UnspecifiedValue)
	    fprintf (fp, " Az=%g", constraint -> ax [Tz]);

	fprintf (fp, "\n");
    }

    constraint -> aux = NULL;
    return 0;
}


/************************************************************************
 * Function:	WriteForce						*
 *									*
 * Description:	Writes a force to the specified stream if the aux	*
 *		pointer matches the mark flag.  The pointer is then	*
 *		cleared.						*
 ************************************************************************/

static int WriteForce (item)
    Item item;
{
    Force force;


    force = (Force) item;


    if (force -> aux == mark_flag) {
	fprintf (fp, "%s", Quote (force -> name));

	if (force -> expr [Fx])
	    fprintf (fp, " Fx=%s", force -> text [Fx]);
	else if (force -> force [Fx])
	    fprintf (fp, " Fx=%g", force -> force [Fx]);

	if (force -> expr [Fy])
	    fprintf (fp, " Fy=%s", force -> text [Fy]);
	else if (force -> force [Fy])
	    fprintf (fp, " Fy=%g", force -> force [Fy]);

	if (force -> expr [Fz])
	    fprintf (fp, " Fz=%s", force -> text [Fz]);
	else if (force -> force [Fz])
	    fprintf (fp, " Fz=%g", force -> force [Fz]);

	if (force -> expr [Mx])
	    fprintf (fp, " Mx=%s", force -> text [Mx]);
	else if (force -> force [Mx])
	    fprintf (fp, " Mx=%g", force -> force [Mx]);

	if (force -> expr [My])
	    fprintf (fp, " My=%s", force -> text [My]);
	else if (force -> force [My])
	    fprintf (fp, " My=%g", force -> force [My]);

	if (force -> expr [Mz])
	    fprintf (fp, " Mz=%s", force -> text [Mz]);
	else if (force -> force [Mz])
	    fprintf (fp, " Mz=%g", force -> force [Mz]);

	fprintf (fp, "\n");
    }

    force -> aux = NULL;
    return 0;
}


/************************************************************************
 * Function:	WriteAnalysisParameters					*
 *									*
 * Description:	Writes out the analysis parameters section.		*
 ************************************************************************/

static void WriteAnalysisParameters ( )
{
    unsigned	 i;
    static char *dof_symbols [ ] = {"", "Tx", "Ty", "Tz", "Rx", "Ry", "Rz"};


    fprintf (fp, "duration=%g ", analysis.duration);
    fprintf (fp, "dt=%g ", analysis.dt);
    fprintf (fp, "beta=%g ", analysis.beta);
    fprintf (fp, "gamma=%g ", analysis.gamma);
    fprintf (fp, "alpha=%g ", analysis.alpha);
 
    if (analysis.mass_mode == 'l')
	fprintf (fp, "mass-mode=lumped");
    else if (analysis.mass_mode == 'c')
	fprintf (fp, "mass-mode=consistent");

    fprintf (fp, "\nnodes=[");
    for (i = 1; i <= analysis.numnodes; i ++)
	if (i < analysis.numnodes)	
	    fprintf (fp, "%d, ", analysis.nodes [i]);
	else
	    fprintf (fp, "%d]\n", analysis.nodes [i]);

    fprintf (fp, "dofs=[");
    for (i = 1; i <= analysis.numdofs; i ++)
	if (i < analysis.numdofs)	
	    fprintf (fp, "%s, ", dof_symbols [(int) analysis.dofs [i]]);
	else
	    fprintf (fp, "%s]\n", dof_symbols [(int) analysis.dofs [i]]);
}


/************************************************************************
 * Function:	WriteFile						*
 *									*
 * Description:	Writes a felt file.  A filename of "-" indicates	*
 *		standard output.					*
 ************************************************************************/

static int WriteFile (filename, flag)
    char *filename;
    char *flag;
{
    int		 any_forces;
    int		 any_loads;
    int		 type;
    Node	 node;
    Element	 element;
    unsigned	 i;
    unsigned	 j;
    static char	*analysis_names [ ] = {"", "static", "transient", "modal"};


    /* Open the file creating it if necessary. */

    if (strcmp (filename, "-")) {
	if (!(fp = fopen (filename, "w"))) {
	    error ("Unable to open %s", filename);
	    return 1;
	}
    } else
	fp = stdout;


    /* Write the problem description section. */

    fprintf (fp, "problem description\n");
    if (problem.title != NULL && strcmp (problem.title, ""))   
       fprintf (fp, "title=%s\n", Quote (problem.title));

    fprintf (fp, "nodes=%u ", problem.num_nodes);
    fprintf (fp, "elements=%u", problem.num_elements);

    if (problem.mode != Static)
	fprintf (fp, " analysis=%s", analysis_names [problem.mode]);

    fprintf (fp, "\n");


    /* Write the analysis parameters section if the mode is not static. */

    if (problem.mode != Static) {
        fprintf (fp, "\nanalysis parameters\n");
	WriteAnalysisParameters ( );	
    }


    /* Write the nodes section marking referenced objects. */

    mark_flag = flag;
    any_forces = 0;
    any_loads = 0;

    if (problem.num_nodes) {
	prev_node = NULL;
	fprintf (fp, "\nnodes\n");

	for (i = 1; i <= problem.num_nodes; i ++) {
	    node = problem.nodes [i];
	    WriteNode (node);
	    prev_node = node;

	    node -> constraint -> aux = flag;
	    if (node -> force) {
		node -> force -> aux = flag;
		any_forces = 1;
	    }
	}
    }


    /* Write the elements section marking referenced objects. */

    if (problem.num_elements) {
	prev_element = NULL;

	for (i = 1; i <= problem.num_elements; i ++) {
	    element = problem.elements [i];
	    type = element -> definition -> type;

	    if (!prev_element || prev_element -> definition -> type != type)
		fprintf(fp,"\n%s elements\n", Quote (ElementArray [type].name));

	    WriteElement (element);
	    prev_element = element;

	    element -> material -> aux = flag;
	    if (element -> numdistributed)
		any_loads = 1;

	    for (j = 1; j <= element -> numdistributed; j ++)
		element -> distributed [j] -> aux = flag;
	}
    }


    /* Write the materials section. */

    if (TreeSize (problem.material_tree) > 0) {
	fprintf (fp, "\nmaterial properties\n");
	TreeSetIterator (problem.material_tree, WriteMaterial);
	TreeIterate (problem.material_tree);
    }


    /* Write the distributed loads section. */

    if (any_loads || !mark_flag) {
	fprintf (fp, "\ndistributed loads\n");
	TreeSetIterator (problem.distributed_tree, WriteLoad);
	TreeIterate (problem.distributed_tree);
    }


    /* Write the constraints section. */

    if (TreeSize (problem.constraint_tree) > 0) {
	fprintf (fp, "\nconstraints\n");
	TreeSetIterator (problem.constraint_tree, WriteConstraint);
	TreeIterate (problem.constraint_tree);
    }


    /* Write the forces section. */

    if (any_forces || !mark_flag) {
	fprintf (fp, "\nforces\n");
	TreeSetIterator (problem.force_tree, WriteForce);
	TreeIterate (problem.force_tree);
    }


    /* Clean up, close the file, and return success. */

    fprintf (fp, "\nend\n");

    if (fp != stdout)
	fclose (fp);

    return 0;
}


/************************************************************************
 * Function:	WriteFeltFile						*
 *									*
 * Description:	Writes a felt file -- only referenced objects will be	*
 *		written.						*
 ************************************************************************/

int WriteFeltFile (filename)
    char *filename;
{
    return WriteFile (filename, (char *) 1);
}


/************************************************************************
 * Function:	DumpFeltFile						*
 *									*
 * Description:	Dumps a felt file -- referenced and unreferenced	*
 *		objects will be written.				*
 ************************************************************************/

int DumpFeltFile (filename)
    char *filename;
{
    return WriteFile (filename, (char *) 0);
}
